//////////////////////////////////////////////////////////////////////////////////////////
///////	Audio looping and mixing /////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cardfeed.h"
#include "loopdata.h"
#include "gui.h"
#include "audio.h"
#include "fxdata.h"

char * looper_output[NUM_LOOPS];
char * looper_effected[NUM_LOOPS];

unsigned int samples_counter = 0;

int left_vu = 0; int right_vu = 0;

void init_effect_buffers()
{
	int L, C, T;
	for (L=0;L<NUM_LOOPS;L++)
		for (C=0;C<FX_ROW;C++)
			for (T=0;T<2;T++)
	{	// mono sized buffers for effects //
		fx_layout[L][C].in_buf[T] = malloc(buf_size/2);
		fx_layout[L][C].out_buf[T] = malloc(buf_size/2);
		
	}
	
}


void mixdown_looper()
{
	int nSamps = buf_size / 4;
	int n, i;
	int left_tmp = 0, right_tmp = 0;
	int left_factor, right_factor;
	int left_total, right_total;
	int left_max, right_max;
	left_max = 0; right_max = 0;
	for (n=0;n<nSamps;n++)
	{
		left_total = right_total = 0;
		for (i=0;i<NUM_LOOPS;i++)
		{
			left_tmp = right_tmp = * ( n + (short *) looper_output[i]); 
			left_factor = right_factor = scr1_parms[4][i];
			left_factor *= scr1_parms[5][i];
			right_factor *= (1000 - scr1_parms[5][i]);
			left_factor /= 1000; right_factor /= 1000;
			left_tmp *= left_factor; left_tmp /= 1000;
			right_tmp *= right_factor; right_tmp /= 1000;
			left_total += left_tmp; right_total += right_tmp;
		}
		left_total = (left_total < -32767)? -32767 : (left_total > 32767) ? 32767 : left_total;
		right_total = (right_total < -32767) ? -32767 :(right_total > 32767) ? 32767: right_total;
		*  ( (short *) buf + n*2) = left_total;
		*  ( (short *) buf + n*2 + 1) = right_total;
		if ( abs(left_total) > left_max) left_max = abs(left_total);
		if ( abs(right_total) > right_max) right_max = abs(right_total);
	}
	left_vu = left_max / (32768 / 128); right_vu = right_max / (32768 / 128);
}

void patch_effect_output()
{
	int L;
	for (L=0;L<NUM_LOOPS;L++)
	{
		if (scr1_fx[L])
			memcpy (looper_output[L], looper_effected[L], buf_size/2);
	}
	
}


void run_looper()
{
	int i, n;
	int measure_factor;
	int nSamps = buf_size / 4; 
	short * loops[NUM_LOOPS];
	int loop_len[NUM_LOOPS];
	int l_start, l_end, tmp , l_pos, l_phased, l_phase_offset;
	float pstart, pend, pphase, f_pos, f_step, f_phase_step, f_phase_offset;

	static int run_looper_init = 1;
	int BPM_ticks;
	
	BPM_ticks = 44100 * 60 / get_bpm();
	if (run_looper_init)
	{
		run_looper_init=0;	// smaller because it's only mono //
		for (i=0;i<NUM_LOOPS;i++) looper_output[i] = malloc(buf_size/2);
		for (i=0;i<NUM_LOOPS;i++) looper_effected[i] = malloc(buf_size/2);
	}
	for (i=0;i<NUM_LOOPS;i++) 
		{loops[i] = (short *)loopData[i];loop_len[i] = loopLength[i]/2;}
	for (i=0; i < NUM_LOOPS; i++)
	{
		memset(looper_output[i], 0, buf_size/2);
		if (isLoopLocked(i)) continue; // not using this loop!
		if (!isLoopFull(i)) continue;
		pstart = (float)scr1_parms[1][i] * 0.001f;
		pend = (float)scr1_parms[2][i] * 0.001f;
		l_start = (int)  ( (float) loop_len[i] * pstart);
		l_end = (int) ( (float) loop_len[i] * pend);
		l_start = (l_start < 0) ? 0 : (l_start >= loop_len[i]) ? loop_len[i]-1 : l_start;
		l_end = (l_end < 0 ) ? 0 : (l_end >= loop_len[i]) ? loop_len[i] - 1: l_end;
		if (l_start > l_end) {tmp = l_start; l_start = l_end; l_end = tmp;}
		if (l_start == l_end) continue; // get the H out of here!
		pphase = (float) scr1_parms[3][i] * 0.001f;
		
		l_phase_offset = (int) ( (float) (l_end - l_start) * pphase);
		f_phase_offset = (float) loopLastPhaseOffset[i];
		f_phase_step = (f_phase_offset - (float) l_phase_offset) / nSamps;
		l_pos = loopPos[i]/2;
		f_pos = (float) l_pos;
		f_step = (float)scr1_parms[0][i] * 0.01f;
		// allow negative speed now ! /// //if (f_step < 0.0f ) f_step = 0.0f;
		l_phased = 0;
		measure_factor = 1 << scr1_measures[i];
		
		for (n = 0; n < nSamps; n++)
		{	
			if ( 0 == ((n + samples_counter) % (measure_factor*BPM_ticks/4)))
			{
				l_pos = l_start;
				loopPos[i] = l_pos * 2;
				f_pos = (float) l_pos;
			}
			l_phased = (int)f_phase_offset + l_pos;
			while (l_phased >= l_end) l_phased -= l_end - l_start;
			while (l_phased < 0) l_phased += l_end;
			*( n + (short *)looper_output[i] ) = *( loops[i] + l_phased);
			// update phase blend...
			f_phase_offset += f_phase_step;
			f_pos += f_step;
			// wrap around start and end
			l_pos = (int) f_pos;
			if (f_step > 0.0f)
			{
				if (l_pos >= l_end) 
					{ l_pos = l_start; f_pos = (float) l_pos;}
			}
			else   if (l_pos < l_start)
				{ l_pos = l_end - 1; f_pos = (float) l_pos; }	
		}
		loopPos[i] = l_pos * 2;
		loopPhasedPos[i] = l_phased;
		loopLastPhaseOffset[i] = (int) ( (float) (l_end - l_start) * pphase);
	}
	samples_counter += nSamps;
}

static short * left_input = 0;
static short * right_input = 0;

void record_loopers()
{
	int nSamps, n; 
	short * sbuf;
	if (left_input == 0 || right_input == 0) 
	{
		left_input = malloc(buf_size / 2);
		right_input = malloc(buf_size / 2);
	}
	sbuf = (short *) buf;
	nSamps = buf_size / 4;
	for (n=0;n<nSamps; n++)
	{
		left_input[n] = *sbuf; sbuf++;
		right_input[n] = *sbuf; sbuf++;
	}
	
	for (n=0;n<NUM_LOOPS;n++) record_looper(n);
}


void record_looper(int L)
{
	int nSamps, n, i, i_dir;
	int BPM_ticks;
	int measure_factor;	
	int loop_len; short * loop;
	int l_start, l_end, tmp , l_pos, new_l_pos;
	float pstart, pend, f_pos, f_step;
	BPM_ticks = 44100 * 60 / get_bpm();
	nSamps = buf_size / 4;
	// borrowed from run_looper above.  main differences here are that
	// PHASE offset is not present and that, as it is reversed from
	// run_looper in that it copies into the loop, not out, it has to
	// be careful to avoid gaps in the sound at high speeds (which is
	// why I don't feel like trying the PHASE blending and stuff )
	if (scr1_record[L] == NO_RECORD) return; // not recording here today!
	if (!isLoopFull(L)) return;  // no loop in this slot to write into!
	if (isLoopLocked(L)) return; // can't touch the loop!

	loop_len = loopLength[L] / 2; loop = (short *) loopData[L];
	
	pstart = (float)scr1_parms[1][L] * 0.001f;
	pend = (float)scr1_parms[2][L] * 0.001f;
	l_start = (int)  ( (float) loop_len * pstart);
	l_end = (int) ( (float) loop_len * pend);
	l_start = (l_start < 0) ? 0 : (l_start >= loop_len) ? loop_len - 1 : l_start;
	l_end = (l_end < 0 ) ? 0 : (l_end >= loop_len) ? loop_len - 1: l_end;
	if (l_start > l_end) {tmp = l_start; l_start = l_end; l_end = tmp;}
	if (l_start == l_end) return; // nowhere to write to!
	l_pos = loopPos[L]/2;
	f_pos = (float) l_pos;
	f_step = (float)scr1_parms[0][L] * 0.01f;

	measure_factor = 1 << scr1_measures[L];	
		
	for (n = 0; n < nSamps; n++)
	{	
		if ( 0 == ((n + samples_counter) % (measure_factor*BPM_ticks/4)))
			{
				l_pos = l_start;
				f_pos = (float) l_pos;
			}
			
			// wrap around start and end
			f_pos += f_step;
			new_l_pos = (int) f_pos;
			
			if (f_step > 0.0f)
			{
				if (new_l_pos >= l_end) 
					{ new_l_pos = l_start; f_pos = (float) new_l_pos;}
			}
			else   if (new_l_pos < l_start)
				{ new_l_pos = l_end - 1; f_pos = (float) new_l_pos; }	
	
			// okay, now l_pos is where we are in the loop we're going into,
			// new_l_pos is the place where l_pos will be afterwords.
			// fill the current input into l_pos and new_l_pos.
			// go in the direction of the speed!
			i = l_pos; i_dir = (f_step > 0.0f ? 1 : -1);
			while (i != new_l_pos)
			{
				switch (scr1_record[L])
				{
				case L_DUB:
					tmp = left_input[n] + loop [i];
					tmp = (tmp < -32767) ? -32767 :
							(tmp > 32767) ? 32767 : tmp;
					loop[i] = tmp; break;
				case R_DUB:
					tmp = right_input[n] + loop [i];
					tmp = (tmp < -32767) ? -32767 :
							(tmp > 32767) ? 32767 : tmp;
					loop[i] = tmp; break;
				case L_RECORD: loop[i] = left_input[n]; break;
				case R_RECORD: loop[i] = right_input[n]; break;
				}
				
				i+= i_dir; 
				if (i_dir > 0 )
				{
					if (i >= l_end) i = l_start;
				}
				else if (i < l_start) i = (l_end - 1);
				
			};
			
	
			l_pos = new_l_pos;
	}	
}

// the heart and soul of ELE.  gets data out of looper_outputs[], runs it through all the effects,
// and puts it into looper_effected[] ////////////////////////////////////////////////////////////

char DontClearMe[FX_ROW*NUM_LOOPS*2];

inline void DCM(int L, int C, int T)
{
	if (C >= FX_ROW) return; // output bus! 
	DontClearMe[(L + C*NUM_LOOPS)*2 + T] = 1;
}

inline int isDCM (int L, int C, int T)
{
	return DontClearMe[(L + C*NUM_LOOPS)*2 + T];
}

void run_effects()
{
	int L, C, T;
	PumpFunc P;
	effect_stat * es;
	effect_type * et;
	short * exchange;
	// check for mutex! important!
	// gives up time slice if effects are busy
	// a glitch is better than a crash!
	fx_lock();
	// begin by clearing all looper_effected outputs
	for (L=0;L<NUM_LOOPS;L++) memset(looper_effected[L], 0, buf_size/2);
	// also clear un-connected inputs of existing effects
	
	memset(DontClearMe, 0, FX_ROW*NUM_LOOPS*2);
	
	for (L=0;L<NUM_LOOPS;L++)
		for (C=0;C<FX_ROW;C++)
	{
		es = (effect_stat *) & fx_layout[L][C];
		if (es->type != -1)
		{
			if (es->link_is[0])
				DCM(	es->link_Loop[0],
					es->link_Col[0],
					es->link_Term[0]);
			if (es->link_is[1])
				DCM(	es->link_Loop[1],
					es->link_Col[1],
					es->link_Term[1]);
 
		}	
		
	}
	
	for (L=0;L<NUM_LOOPS;L++)
		for (C=0;C<FX_ROW;C++)
	{
		es = (effect_stat *) & fx_layout[L][C];
		if (es->type != -1)
			for (T=0;T<2;T++)
				if (!isDCM(L, C, T)) memset (es->in_buf[T], 0, buf_size/2);	
	}
	
	
	// now feed looper_output into effect buffers
	for (L=0;L<NUM_LOOPS;L++)
	{
		if (loopbus_links[L].is_linked)
		{
			if (loopbus_links[L].Col == FX_ROW) // straight connect to output
			{
				memcpy(looper_effected[loopbus_links[L].Loop], 
					looper_output[L]
					, buf_size/2);
			}
			else
			{
				es = (effect_stat *) & fx_layout
					[loopbus_links[L].Loop]
					[loopbus_links[L].Col];
				memcpy(es->in_buf[loopbus_links[L].Term], looper_output[L],
					buf_size/2);
			}
			
			
		}
		
	}
	// this bird's too small to feed us all!
	// so pump it up, dawg, pump up the turkey, pump it up
	for (C=0;C<FX_ROW;C++)
		for (L=0;L<NUM_LOOPS;L++)
	{
		es = (effect_stat *) & fx_layout[L][C];
		if (es->type == -1) continue; // no effect here!
		et = (effect_type *) & effect_types[es->type];
		
		// shall we run a function?
		if (et->func != NULL)
		{
			P = (PumpFunc) et->func;
			(*P)(L, C);
			// give it up!
			for (T=0;T<2;T++)
				if (es->link_is[T])
			{
				// send it to an output?
				if (es->link_Col[T] >= FX_ROW)
				{
					memcpy(looper_effected[es->link_Loop[T]],
						es->out_buf[T],
						buf_size/2);
				}
				else
				{
				/*
					memcpy(fx_layout
					[es->link_Loop[T]]
					[es->link_Col[T] ].in_buf[es->link_Term[T]]
					,es->out_buf[T]
					,buf_size/2);
				*/
				// try for a little more efficiency than copy here //
				exchange = fx_layout[es->link_Loop[T]][es->link_Col[T]].in_buf
						[es->link_Term[T]];
				fx_layout[es->link_Loop[T]][es->link_Col[T]].in_buf
						[es->link_Term[T]] = es->out_buf[T];
				es->out_buf[T] = exchange;
				}
			}
			
				
		}
		
		
	}
	// let go of the darned mutex
	fx_unlock();
}

