////////////////////////////////////////////
///     Handles keyboard and mouse input ///
////////////////////////////////////////////

#include <stdio.h>
#include <vga.h>
#include <vgakeyboard.h>
#include <vgamouse.h>
#include <string.h>
#include <termios.h>
#include <term.h>
#include <unistd.h>
#include "input.h"
#include "loopdata.h"

static struct termios terminal_settings;

char kdown[256];
char kpush[256];
// tick value when key was pushed down //
unsigned int ktime[256];
static int is_slow = 0;
void slow_keys(int mode)
{
	is_slow = 0;
	if (mode) is_slow = 1;
}


void init_input()
{
	keyboard_init();
	vga_setmousesupport(1);
	memset(kdown,0,256);
	memset(kpush,0,256);

}

void close_input()
{
	// keyboard_close();
	// mouse_close();
}

extern volatile unsigned int ticks;

void scan_mouse()
{
	int MDX, MDY, ms, mv;
	int mx, my, mz, mrx, mry, mrz;
	mouse_setrange_6d (0, 0, 0, 0, 0, 0,
				-180, 180, -180, 180, -180, 180,
				MOUSE_RXDIM|MOUSE_RYDIM|MOUSE_RZDIM);
	mouse_setposition_6d(200,200, 0,0,0,0, MOUSE_6DIM);
	if (!mouse_update()) return;
	mouse_getposition_6d(&mx, &my, &mz, &mrx, &mry, &mrz);
	MDX = mx - 200;
	MDY = my - 200;
	MDY = - MDY;
	for (ms=0;ms<8;ms++)
	{
	mv = mouse_spinner_value[ms];
	switch (mouse_spinner_axis[ms])
	{
	default: case NO_AXIS: break; 
	case X_AXIS: mv += MDX; break;
	case Y_AXIS: mv += MDY; break;
	case MINUS_X_AXIS: mv -= MDX; break;
	case MINUS_Y_AXIS: mv -= MDY; break;
	case X2_AXIS: mv += 2*MDX; break;
	case Y2_AXIS: mv += 2*MDY; break;
	case MINUS_X2_AXIS: mv -= 2*MDX; break;
	case MINUS_Y2_AXIS: mv -= 2*MDY; break;
	case WHEEL1: mv -= mrx; break;
	case WHEEL2: mv -= mry; break;
	case WHEEL3: mv -= mrz; break;
	case MINUS_WHEEL1: mv += mrx; break;
	case MINUS_WHEEL2: mv += mry; break;
	case MINUS_WHEEL3: mv += mrz; break;
	};
	mv = (mv < 0) ? 0 : (mv > 999) ? 999 : mv;
	mouse_spinner_value[ms]=mv;
	}

}


void scan_keys()
{
	char * km; int i;
	int kKEY_REPEAT, kKEY_REREPEAT;
	keyboard_update();
	km = keyboard_getstate();
	kKEY_REPEAT = KEY_REPEAT;
	kKEY_REREPEAT = KEY_REREPEAT;
	if (is_slow)
	{
		kKEY_REPEAT *= 10;
		kKEY_REREPEAT *= 10;
	}
	
	// set kpush if key is newly pressed //
	for (i=0;i<256;i++)
		if (km[i]==KEY_PRESSED && 
			(kdown[i]!=KEY_PRESSED ||
				(i == SCANCODE_RIGHTSHIFT ||
				 i == SCANCODE_LEFTSHIFT) ) ) 
		{
			kpush[i]=KEY_PRESSED;
			ktime[i]=ticks;
		}
		else	
		{	
			if (km[i]==KEY_PRESSED )
			{	// ESCAPE doesn't repeat
				if ((ticks-ktime[i])>kKEY_REPEAT && i != SCANCODE_ESCAPE)
				{
					ktime[i]=ticks;
					ktime[i]+=kKEY_REPEAT;
					ktime[i]-=kKEY_REREPEAT;
					kpush[i]=KEY_PRESSED;
				}
				else kpush[i]=0;
			}
			else
				kpush[i]=0;
		}
	memcpy(kdown, km, 256);
	
	
}
	
void save_terminal_settings()
{
	tcgetattr(0, &terminal_settings);
}

void restore_terminal_settings()
{
	tcsetattr(0, TCSANOW, &terminal_settings);
}

